import json
from typing import Dict, List, Optional, Union

from django.db.models import (
    DateField,
    Field,
    ForeignKey,
    ManyToManyField,
    ManyToManyRel,
    ManyToOneRel,
    Model,
    OneToOneField,
    OneToOneRel,
    TimeField,
)
from django.db.models.fields import AutoFieldMixin
from forestadmin.agent_toolkit.forest_logger import ForestLogger
from forestadmin.datasource_django.utils.type_converter import FilterOperator, TypeConverter
from forestadmin.datasource_toolkit.interfaces.fields import (
    Column,
    FieldType,
    ManyToMany,
    ManyToOne,
    OneToMany,
    OneToOne,
    Operator,
    PolymorphicManyToOne,
    PolymorphicOneToMany,
    PolymorphicOneToOne,
    PrimitiveType,
    Validation,
)
from forestadmin.datasource_toolkit.interfaces.models.collections import CollectionSchema
from typing_extensions import TypeGuard

try:
    # GeneratedField is available since django 5
    from django.db.models import GeneratedField
except ImportError:
    GeneratedField = None


class FieldFactory:
    @staticmethod
    def _build_enum_values(field: Field) -> Optional[List[str]]:
        if field.choices:
            return [str(c[0]) for c in field.choices]  # type: ignore
        return None

    @staticmethod
    def _build_is_read_only(field: Field) -> bool:
        if field.primary_key:
            return isinstance(field, AutoFieldMixin)
        elif isinstance(field, DateField) or isinstance(field, TimeField):
            return field.auto_now is True or field.auto_now_add is True
        if GeneratedField is not None and isinstance(field, GeneratedField):
            return True
        return not field.editable

    @classmethod
    def _build_validations(cls, field: Field) -> List[Validation]:
        validations: List[Validation] = []
        if not (field.null or field.blank) and not cls._build_is_read_only(field) and not field.has_default():
            validations.append(
                {
                    "operator": Operator.PRESENT,
                }
            )
        if field.max_length:
            validations.append({"operator": Operator.SHORTER_THAN, "value": field.max_length})

        return validations

    @classmethod
    def build(cls, field: Field, current_model: Model) -> Column:
        column_type = TypeConverter.convert(field)  # type: ignore

        default_value = field.default
        try:
            # field.get_default() evaluate a function if default is a function
            # we prefer don't have a default value rather than a false one
            json.dumps(default_value)
        except TypeError:  # not JSON Serializable
            default_value = None

        # when model inheritance (https://docs.djangoproject.com/en/5.0/topics/db/models/#multi-table-inheritance),
        # you inherit from parent model attributes, and also his pk ;
        # but the pk of the parent model is not the pk of the current model!
        is_primary_key = field.primary_key and current_model == field.model

        return {
            "column_type": column_type,
            "is_primary_key": is_primary_key,  # type: ignore
            "is_read_only": cls._build_is_read_only(field),
            "default_value": default_value,
            "is_sortable": True,
            "validations": cls._build_validations(field),
            "filter_operators": FilterOperator.get_operators_for_field(field),
            "enum_values": cls._build_enum_values(field),
            "type": FieldType.COLUMN,
        }


class DjangoCollectionFactory:
    @staticmethod
    def _build_one_to_many(relation: ManyToOneRel) -> OneToMany:
        return {
            "foreign_collection": relation.target_field.model._meta.db_table,
            "origin_key": relation.field.attname,
            "origin_key_target": relation.field.target_field.attname,
            "type": FieldType.ONE_TO_MANY,
        }

    @staticmethod
    def _build_many_to_one(relation: Union[OneToOneField, ForeignKey, ManyToOneRel]) -> ManyToOne:
        if isinstance(relation, ManyToOneRel):
            foreign_key = relation.field.attname
        elif isinstance(relation, ForeignKey) or isinstance(relation, OneToOneField):
            foreign_key = relation.attname
        return {
            "foreign_collection": relation.target_field.model._meta.db_table,
            "foreign_key": foreign_key,
            "foreign_key_target": relation.target_field.attname,
            "type": FieldType.MANY_TO_ONE,
        }

    @staticmethod
    def _build_one_to_polymorphic(
        relation: "GenericRelation", model: Model  # noqa: F821  # type: ignore
    ) -> Union[PolymorphicOneToMany, PolymorphicOneToOne]:
        foreign_model = relation.target_field.model
        ret = {
            "foreign_collection": foreign_model._meta.db_table,  # app_addresses
            "origin_key": relation.object_id_field_name,  # addressable_id
            "origin_key_target": model._meta.pk.name,  # id
            "origin_type_field": relation.content_type_field_name,  # addressable_type
            "origin_type_value": model._meta.db_table,  # app_customer
        }

        for uniqueness in relation.target_field.model._meta.unique_together:
            if len(uniqueness) == 2 and ret["origin_type_field"] in uniqueness and ret["origin_key"] in uniqueness:
                ret["type"] = FieldType.POLYMORPHIC_ONE_TO_ONE
                break
        else:
            ret["type"] = FieldType.POLYMORPHIC_ONE_TO_MANY
        return ret  # type:ignore

    @staticmethod
    def _build_field_override_for_many_to_one_polymorphic(
        relation: "GenericForeignKey", model: Model, foreign_collections: List[str]  # noqa: F821  # type: ignore
    ) -> Dict[str, Dict]:
        overrides = {}
        # override fk field
        overrides[relation.fk_field] = {
            **FieldFactory.build(model._meta.get_field(relation.fk_field), model),
            "is_read_only": True,
        }
        # override ct field
        type_field = model._meta.get_field(relation.ct_field)
        overrides[relation.ct_field] = {
            "column_type": PrimitiveType.ENUM,
            "is_primary_key": type_field.primary_key,  # type: ignore
            "is_read_only": True,
            "default_value": None,
            "is_sortable": True,
            "validations": [],
            "filter_operators": FilterOperator.get_for_type(PrimitiveType.STRING),
            "enum_values": foreign_collections,
            "type": FieldType.COLUMN,
        }
        # override ct_id field
        overrides[type_field.column] = {
            **FieldFactory.build(model._meta.get_field(type_field.column), model),
            "is_read_only": True,
        }
        return overrides

    @staticmethod
    def _build_many_to_one_polymorphic(
        relation: "GenericForeignKey", model: Model  # noqa: F821  # type: ignore
    ) -> PolymorphicManyToOne:
        targets = {}
        for field in model._meta.get_fields(include_hidden=True):
            if (
                not is_generic_rel(field)
                or field.remote_field.content_type_field_name != relation.ct_field  # type:ignore
                or field.remote_field.object_id_field_name != relation.fk_field  # type:ignore
            ):
                continue
            targets[field.remote_field.model._meta.db_table] = field.remote_field.model._meta.pk.name  # type:ignore
        return {
            "foreign_collections": [*targets.keys()],
            "foreign_key": relation.fk_field,
            "foreign_key_type_field": relation.ct_field,
            "foreign_key_targets": targets,
            "type": FieldType.POLYMORPHIC_MANY_TO_ONE,
        }

    @staticmethod
    def _build_one_to_one(relation: OneToOneRel) -> OneToOne:
        return {
            "foreign_collection": relation.target_field.model._meta.db_table,
            "origin_key": relation.field.attname,
            "origin_key_target": relation.field.target_field.attname,
            "type": FieldType.ONE_TO_ONE,
        }

    @staticmethod
    def _build_many_to_many(relation: Union[ManyToManyField, ManyToManyRel]) -> ManyToMany:
        kwargs: Dict[str, str] = {"foreign_collection": relation.target_field.model._meta.db_table}

        if isinstance(relation, ManyToManyField):
            kwargs["through_collection"] = relation.remote_field.through._meta.db_table

            kwargs["origin_key"] = relation.m2m_column_name()
            kwargs["foreign_key"] = relation.m2m_reverse_name()

        elif isinstance(relation, ManyToManyRel):
            kwargs["through_collection"] = relation.through._meta.db_table

            kwargs["origin_key"] = relation.field.m2m_reverse_name()
            kwargs["foreign_key"] = relation.field.m2m_column_name()

        kwargs["origin_key_target"] = relation.model._meta.pk.attname
        kwargs["foreign_key_target"] = relation.target_field.model._meta.pk.attname

        return ManyToMany(type=FieldType.MANY_TO_MANY, foreign_relation=None, **kwargs)

    @staticmethod
    def build(model: Model, introspect_polymorphic_relations: bool) -> CollectionSchema:  # noqa:C901
        fields = {}
        overrides = {}
        for field in model._meta.get_fields(include_hidden=True):
            if not field.is_relation:
                fields[field.name] = FieldFactory.build(field, model)
            else:
                if is_generic_rel(field):
                    continue  # will be handled during is_generic_foreign_key

                # get_fields with include hidden doesn't include autogenerated foreign key fields (ending with _id)
                if isinstance(field, ForeignKey):
                    fields[field.attname] = FieldFactory.build(field, model)

                if field.one_to_one is True:
                    if isinstance(field, OneToOneField):
                        fields[field.name] = DjangoCollectionFactory._build_many_to_one(field)
                    else:
                        fields[field.name] = DjangoCollectionFactory._build_one_to_one(field)

                elif field.one_to_many is True:
                    if is_generic_relation(field):
                        if introspect_polymorphic_relations:
                            fields[field.name] = DjangoCollectionFactory._build_one_to_polymorphic(field, model)
                        else:
                            ForestLogger.log(
                                "info",
                                f"Ignoring {model._meta.db_table}.{field.name} "
                                "because polymorphic relation is not supported.",
                            )
                    else:
                        fields[f"{field.name}_{field.remote_field.name}"] = DjangoCollectionFactory._build_one_to_many(
                            field
                        )

                elif field.many_to_one is True:
                    if is_generic_foreign_key(field):
                        if introspect_polymorphic_relations:
                            fields[field.name] = DjangoCollectionFactory._build_many_to_one_polymorphic(field, model)
                            overrides.update(
                                DjangoCollectionFactory._build_field_override_for_many_to_one_polymorphic(
                                    field, model, fields[field.name]["foreign_collections"]
                                )
                            )
                        else:
                            ForestLogger.log(
                                "info",
                                f"Ignoring {model._meta.db_table}.{field.name} "
                                "because polymorphic relation is not supported.",
                            )

                    else:
                        fields[field.name] = DjangoCollectionFactory._build_many_to_one(field)

                elif field.many_to_many is True:
                    fields[field.name] = DjangoCollectionFactory._build_many_to_many(field)

        if overrides:
            fields.update(overrides)
        return {"actions": {}, "fields": fields, "searchable": False, "segments": [], "countable": True, "charts": []}


def is_generic_foreign_key(field) -> TypeGuard["GenericForeignKey"]:  # noqa: F821 # type: ignore
    """return true if it is the polymorphism relation"""
    # when imported at top level (before app.Ready):
    # it raise django.core.exceptions.AppRegistryNotReady: Apps aren't loaded yet.
    from django.contrib.contenttypes.fields import GenericForeignKey

    return isinstance(field, GenericForeignKey)


def is_generic_relation(field) -> TypeGuard["GenericRelation"]:  # noqa: F821 # type: ignore
    """return true if it is the reverse polymorphism relation"""
    # when imported at top level (before app.Ready):
    # it raise django.core.exceptions.AppRegistryNotReady: Apps aren't loaded yet.
    from django.contrib.contenttypes.fields import GenericRelation

    return isinstance(field, GenericRelation)


def is_generic_rel(field) -> TypeGuard["GenericRel"]:  # noqa: F821 # type: ignore
    """return true if it is a potential relation of the polymorphism relation"""
    # when imported at top level (before app.Ready):
    # it raise django.core.exceptions.AppRegistryNotReady: Apps aren't loaded yet.
    from django.contrib.contenttypes.fields import GenericRel

    return isinstance(field, GenericRel)



from collections import defaultdict
from datetime import date, datetime
from typing import Any, Dict, List, Optional, Set, Tuple

from django.db import models
from forestadmin.datasource_django.exception import DjangoDatasourceException
from forestadmin.datasource_django.interface import BaseDjangoCollection
from forestadmin.datasource_django.utils.polymorphic_util import DjangoPolymorphismUtil
from forestadmin.datasource_django.utils.type_converter import FilterOperator
from forestadmin.datasource_toolkit.interfaces.fields import (
    is_many_to_one,
    is_one_to_one,
    is_polymorphic_many_to_one,
    is_polymorphic_one_to_one,
)
from forestadmin.datasource_toolkit.interfaces.query.aggregation import (
    AggregateResult,
    Aggregation,
    Aggregator,
    DateOperation,
    PlainAggregationGroup,
)
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.nodes.base import ConditionTree
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.nodes.branch import (
    Aggregator as ConditionTreeAggregator,
)
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.nodes.branch import ConditionTreeBranch
from forestadmin.datasource_toolkit.interfaces.query.condition_tree.nodes.leaf import ConditionTreeLeaf
from forestadmin.datasource_toolkit.interfaces.query.filter.paginated import PaginatedFilter
from forestadmin.datasource_toolkit.interfaces.query.filter.unpaginated import BaseFilter, Filter
from forestadmin.datasource_toolkit.interfaces.query.projections import Projection
from forestadmin.datasource_toolkit.interfaces.records import RecordsDataAlias


class DjangoQueryBuilder:
    AGGREGATION_FUNC_MAPPING = {
        Aggregator.COUNT: models.Count,
        Aggregator.AVG: models.Avg,
        Aggregator.MAX: models.Max,
        Aggregator.MIN: models.Min,
        Aggregator.SUM: models.Sum,
    }

    @classmethod
    def _normalize_projection(cls, projection: Projection, prefix: str = "") -> Projection:
        # needed to be compliant with the orm result orm
        normalized_projection = [f"{prefix}{col}" for col in projection.columns]
        for parent_field, child_fields in projection.relations.items():
            normalized_projection.extend(cls._normalize_projection(child_fields, f"{prefix}{parent_field}__"))
        return Projection(*normalized_projection)

    @classmethod
    def _mk_base_queryset(
        cls,
        collection: BaseDjangoCollection,
        largest_projection: Projection,
        filter_: BaseFilter,
    ) -> models.QuerySet:
        qs: models.QuerySet = collection.model.objects.all()  # type:ignore

        select_related, prefetch_related = DjangoQueryBuilder._find_related_in_projection(
            collection, largest_projection
        )

        with_generic_fk = DjangoPolymorphismUtil.is_polymorphism_implied(largest_projection, collection)
        if with_generic_fk:
            select_related.union(
                [
                    collection.schema["fields"][generic_fk]["foreign_key_type_field"]  # type:ignore
                    for generic_fk in DjangoPolymorphismUtil.get_polymorphism_relations(largest_projection, collection)
                ]
            )

        qs = qs.select_related(*cls._normalize_projection(Projection(*select_related))).prefetch_related(
            *cls._normalize_projection(Projection(*prefetch_related)),
        )

        condition_tree = filter_.condition_tree
        if condition_tree:
            if any(
                [
                    DjangoPolymorphismUtil.is_type_field_of_generic_fk(field, collection)
                    for field in filter_.condition_tree.projection
                ]
            ):
                condition_tree = DjangoPolymorphismUtil.replace_content_type_in_condition_tree(
                    filter_.condition_tree, collection
                )

        qs = qs.filter(DjangoQueryConditionTreeBuilder.build(condition_tree))

        if isinstance(filter_, PaginatedFilter):
            qs = qs.order_by(*DjangoQueryPaginationBuilder.get_order_by(filter_))
        return qs

    @classmethod
    def _find_related_in_projection(
        cls,
        collection: BaseDjangoCollection,
        projection: Projection,
    ) -> Tuple[Set[str], Set[str]]:
        select_related = set()
        prefetch_related = set()
        break_select_related: Dict[str, bool] = defaultdict(lambda: False)

        for relation_name, subfields in projection.relations.items():
            field_schema = collection.schema["fields"][relation_name]
            if not break_select_related[relation_name]:
                if is_many_to_one(field_schema) or is_one_to_one(field_schema):
                    select_related.add(relation_name)
                elif is_polymorphic_many_to_one(field_schema):
                    select_related.add(field_schema["foreign_key_type_field"])
                    break_select_related[relation_name] = True
                    prefetch_related.add(relation_name)
                elif is_polymorphic_one_to_one(field_schema):
                    break_select_related[relation_name] = True
                    prefetch_related.add(relation_name)

                else:
                    break_select_related[relation_name] = True
                    prefetch_related.add(relation_name)

            sub_select, sub_prefetch = (
                cls._find_related_in_projection(
                    collection.datasource.get_collection(field_schema["foreign_collection"]), subfields
                )
                if subfields != ["*"]
                else ([], [])
            )

            if not break_select_related[relation_name]:
                select_related = select_related.union(Projection(*sub_select).nest(relation_name))
            else:
                prefetch_related = prefetch_related.union(Projection(*sub_select).nest(relation_name))
            prefetch_related = prefetch_related.union(Projection(*sub_prefetch).nest(relation_name))

        return select_related, prefetch_related

    @classmethod
    def mk_list(
        cls,
        collection: BaseDjangoCollection,
        filter_: PaginatedFilter,
        projection: Projection,
    ) -> models.QuerySet:
        full_projection = projection
        if filter_.condition_tree:
            full_projection = full_projection.union(filter_.condition_tree.projection)
        if filter_.sort:
            full_projection = full_projection.union(filter_.sort.projection)
        qs = cls._mk_base_queryset(collection, full_projection, filter_)

        # only raise errors when trying to get a field by a to_many relation
        # or in other words if we have something to pass to prefetch_related
        _, prefetch = cls._find_related_in_projection(collection, full_projection)

        if len(prefetch) == 0 and not DjangoPolymorphismUtil.is_polymorphism_implied(projection, collection):
            qs = qs.only(*cls._normalize_projection(full_projection))
        return DjangoQueryPaginationBuilder.paginate_queryset(qs, filter_)

    @classmethod
    def mk_aggregate(
        cls,
        collection: BaseDjangoCollection,
        filter_: Optional[Filter],
        aggregation: Aggregation,
        limit: Optional[int],
    ) -> List[AggregateResult]:
        full_projection = aggregation.projection
        if filter_.condition_tree:
            full_projection = full_projection.union(filter_.condition_tree.projection)
        qs = cls._mk_base_queryset(collection, full_projection, filter_)

        if aggregation.field:
            qs = qs.order_by(f'-{aggregation.field.replace(":", "__")}')

        if len(aggregation.groups) == 0:
            aggregated_field = aggregation.operation.value
            if aggregation.field is None:
                aggregate_kwargs = {aggregated_field: cls.AGGREGATION_FUNC_MAPPING[aggregation.operation]("*")}
            else:
                aggregate_kwargs = {
                    aggregated_field: cls.AGGREGATION_FUNC_MAPPING[aggregation.operation](aggregation.field)
                }

            qs = qs.aggregate(**aggregate_kwargs)
            value = float(qs.get(aggregated_field, 0))
            return [{"value": value, "group": {}}]

        else:
            aggregated_field = None
            if aggregation.field is None:
                aggregated_field = aggregation.operation.value
                annotate_kwargs = {aggregated_field: cls.AGGREGATION_FUNC_MAPPING[aggregation.operation]("*")}

            else:
                aggregated_field = aggregation.field.replace(":", "__")
                annotate_kwargs = {
                    aggregated_field: cls.AGGREGATION_FUNC_MAPPING[aggregation.operation](aggregated_field)
                }

            fields = DjangoQueryGroupByHelper.mk_groupby_for_values(aggregation.groups)
            qs = qs.values(*fields)
            qs = qs.annotate(**annotate_kwargs)
            if limit:
                qs = qs[0:limit]

            return [DjangoQueryGroupByHelper.parse_groupby_row(row, aggregation.groups, aggregated_field) for row in qs]

    @staticmethod
    def mk_create(collection: BaseDjangoCollection, data: List[RecordsDataAlias]) -> List[models.Model]:
        instances: List[models.Model] = [
            collection.model.objects.create(**DjangoPolymorphismUtil.replace_content_type_in_patch(d, collection))
            for d in data
        ]
        return instances

    @staticmethod
    def mk_update(
        collection: BaseDjangoCollection,
        filter_: Optional[Filter],
        patch: RecordsDataAlias,
    ):
        patch = DjangoPolymorphismUtil.replace_content_type_in_patch(patch, collection)
        qs = collection.model.objects.filter(
            DjangoQueryConditionTreeBuilder.build(
                DjangoPolymorphismUtil.replace_content_type_in_condition_tree(filter_.condition_tree, collection)
            )
        )
        qs.update(**{k.replace(":", "__"): v for k, v in patch.items()})

    @staticmethod
    def mk_delete(collection: BaseDjangoCollection, filter_: Optional[Filter]):
        qs = collection.model.objects.filter(
            DjangoQueryConditionTreeBuilder.build(
                DjangoPolymorphismUtil.replace_content_type_in_condition_tree(filter_.condition_tree, collection)
            )
        )
        qs.delete()


class DjangoQueryConditionTreeBuilder:
    @classmethod
    def _build_leaf_condition(cls, leaf: ConditionTreeLeaf) -> models.Q:
        field = leaf.field.replace(":", "__")
        key, should_negate = FilterOperator.get_operator(leaf.operator)

        value = leaf.value
        if key == "__isnull":
            value = True

        if key == "__in" and isinstance(value, list) and None in value:
            q_obj = cls.build(ConditionTreeBranch("or", [ConditionTreeLeaf(leaf.field, "equal", v) for v in value]))
            return ~q_obj if should_negate else q_obj

        if should_negate:
            return ~models.Q(**{f"{field}{key}": value})
        else:
            return models.Q(**{f"{field}{key}": value})

    @classmethod
    def _aggregate(cls, aggregator: ConditionTreeAggregator, conditions: List[ConditionTree]) -> models.Q:
        ret: models.Q = models.Q()
        if aggregator == ConditionTreeAggregator.AND:
            for cond in conditions:
                ret &= cond
        elif aggregator == ConditionTreeAggregator.OR:
            for cond in conditions:
                ret |= cond
        else:
            raise DjangoDatasourceException(f"Unable to handle the condition tree aggregator {aggregator}")
        return ret

    @classmethod
    def _build_branch_condition(cls, branch: ConditionTreeBranch) -> models.Q:
        conditions = [DjangoQueryConditionTreeBuilder.build(condition) for condition in branch.conditions]
        return DjangoQueryConditionTreeBuilder._aggregate(branch.aggregator, conditions)

    @classmethod
    def build(cls, condition_tree: Optional[ConditionTree]) -> models.Q:
        if condition_tree is None:
            return models.Q()
        if isinstance(condition_tree, ConditionTreeLeaf):
            return cls._build_leaf_condition(condition_tree)
        elif isinstance(condition_tree, ConditionTreeBranch):
            return cls._build_branch_condition(condition_tree)
        raise DjangoDatasourceException(f"Unable to handle ConditionTree type {condition_tree.__class__}")


class DjangoQueryPaginationBuilder:
    @classmethod
    def get_order_by(cls, filter_: PaginatedFilter) -> List[str]:
        ret = []
        if filter_.sort:
            for sorting in filter_.sort:
                if sorting["ascending"]:
                    ret.append(sorting["field"].replace(":", "__"))
                else:
                    ret.append(f"-{sorting['field'].replace(':', '__')}")
        return ret

    @classmethod
    def paginate_queryset(cls, queryset: models.QuerySet, filter_: PaginatedFilter) -> models.QuerySet:
        if filter_.page:
            return queryset[filter_.page.skip : filter_.page.skip + filter_.page.limit]
        return queryset


class DjangoQueryGroupByHelper:
    DATE_OPERATION_SUFFIX_MAPPING: Dict[DateOperation, str] = {
        DateOperation.DAY: "__day",
        DateOperation.WEEK: "__week",
        DateOperation.MONTH: "__month",
        DateOperation.YEAR: "__year",
    }

    @classmethod
    def get_operation_suffixes(cls, group: PlainAggregationGroup) -> List[str]:
        if not group.get("operation"):
            return None

        if group["operation"] == DateOperation.YEAR:
            return [cls.DATE_OPERATION_SUFFIX_MAPPING[DateOperation.YEAR]]
        if group["operation"] == DateOperation.MONTH:
            return [
                cls.DATE_OPERATION_SUFFIX_MAPPING[DateOperation.YEAR],
                cls.DATE_OPERATION_SUFFIX_MAPPING[DateOperation.MONTH],
            ]
        if group["operation"] == DateOperation.WEEK:
            return [
                cls.DATE_OPERATION_SUFFIX_MAPPING[DateOperation.YEAR],
                cls.DATE_OPERATION_SUFFIX_MAPPING[DateOperation.WEEK],
            ]
        if group["operation"] == DateOperation.DAY:
            return [
                cls.DATE_OPERATION_SUFFIX_MAPPING[DateOperation.YEAR],
                cls.DATE_OPERATION_SUFFIX_MAPPING[DateOperation.MONTH],
                cls.DATE_OPERATION_SUFFIX_MAPPING[DateOperation.DAY],
            ]

    @classmethod
    def mk_groupby_for_values(cls, groups: List[PlainAggregationGroup]) -> List[str]:
        values = []
        for group in groups:
            value = group["field"].replace(":", "__")

            suffixes = cls.get_operation_suffixes(group)
            if suffixes:
                values.extend([f"{value}{suffix}" for suffix in suffixes])

            else:
                values.append(value)
        return values

    @classmethod
    def parse_groupby_row(
        cls, row: Dict[str, Any], groups: List[PlainAggregationGroup], aggregated_field: str
    ) -> AggregateResult:
        groupby = {}
        for group in groups:
            suffixes = cls.get_operation_suffixes(group)
            field_name = group["field"].replace(":", "__")
            if not suffixes:
                groupby[group["field"]] = row[field_name]
            else:
                groupby[group["field"]] = cls._make_date_from_record(row, field_name, group["operation"])

        return {"value": float(row[aggregated_field]), "group": groupby}

    @classmethod
    def _make_date_from_record(cls, row: AggregateResult, date_field: str, date_operation: DateOperation) -> date:
        if date_operation == DateOperation.YEAR:
            return date(row[f"{date_field}__year"], 1, 1)

        if date_operation == DateOperation.MONTH:
            return date(row[f"{date_field}__year"], row[f"{date_field}__month"], 1)

        if date_operation == DateOperation.WEEK:
            str_year_week = f'{row[f"{date_field}__year"]}-W{row[f"{date_field}__week"]}'
            row_date = datetime.strptime(str_year_week + "-1", "%Y-W%W-%w")
            return row_date.date()

        if date_operation == DateOperation.DAY:
            return date(row[f"{date_field}__year"], row[f"{date_field}__month"], row[f"{date_field}__day"])



import abc

from django.db.models import Model
from forestadmin.datasource_toolkit.collections import Collection
from forestadmin.datasource_toolkit.datasources import Datasource


class BaseDjangoCollection(Collection, abc.ABC):
    def model(self) -> Model:
        """return model of the collection"""


class BaseDjangoDatasource(Datasource, abc.ABC):
    pass



from typing import Dict, Set, Tuple

import django
from django.db import models
from forestadmin.datasource_django.exception import DjangoDatasourceException
from forestadmin.datasource_toolkit.interfaces.fields import ColumnAlias, Operator, PrimitiveType
from forestadmin.datasource_toolkit.utils.operators import BaseFilterOperator

try:
    # if postgres driver is installed
    from django.contrib.postgres import fields as postgres_fields
except ImportError:
    postgres_fields = None

try:
    # GeneratedField is available since django 5
    from django.db.models import GeneratedField
except ImportError:
    GeneratedField = None


class ConverterException(DjangoDatasourceException):
    pass


if postgres_fields is not None:
    POSTGRES_TYPE: Dict[type, PrimitiveType] = {
        # specific postgres fields
        postgres_fields.HStoreField: PrimitiveType.JSON,
        # postgres_fields.RangeField and subclassed ones not handles
    }
    if django.VERSION[0] < 5 or (django.VERSION[0] == 5 and django.VERSION[1] < 1):
        POSTGRES_TYPE[postgres_fields.CIText] = PrimitiveType.STRING
        POSTGRES_TYPE[postgres_fields.CIEmailField] = PrimitiveType.STRING
else:
    POSTGRES_TYPE: Dict[type, PrimitiveType] = {}


class TypeConverter:
    TYPES: Dict[type, PrimitiveType] = {
        # string
        models.CharField: PrimitiveType.STRING,
        models.TextField: PrimitiveType.STRING,
        #
        # number
        models.IntegerField: PrimitiveType.NUMBER,
        models.FloatField: PrimitiveType.NUMBER,
        models.DecimalField: PrimitiveType.NUMBER,
        #
        # bool
        models.BooleanField: PrimitiveType.BOOLEAN,
        #
        # datetime
        # DateTimeField should be in first because it overrides DateField
        models.DateTimeField: PrimitiveType.DATE,
        models.DateField: PrimitiveType.DATE_ONLY,
        models.TimeField: PrimitiveType.TIME_ONLY,
        # https://docs.djangoproject.com/en/4.2/ref/models/fields/#durationfield
        models.DurationField: PrimitiveType.NUMBER,
        #
        # binary
        models.BinaryField: PrimitiveType.BINARY,
        #
        # file field
        models.FileField: PrimitiveType.STRING,
        models.FilePathField: PrimitiveType.STRING,
        #
        # ip address
        models.IPAddressField: PrimitiveType.STRING,
        models.GenericIPAddressField: PrimitiveType.STRING,
        #
        # json
        models.JSONField: PrimitiveType.JSON,
        #
        # uuid
        models.UUIDField: PrimitiveType.UUID,
        #
        **POSTGRES_TYPE,
    }

    @classmethod
    def convert(cls, field: models.Field) -> ColumnAlias:
        if field.choices is not None:
            return PrimitiveType.ENUM

        if isinstance(field, models.ForeignKey):
            return cls.convert(field.target_field)

        if GeneratedField is not None and isinstance(field, GeneratedField):
            return cls.convert(field.output_field)

        if field.__class__ in cls.TYPES:
            return cls.TYPES[field.__class__]

        if postgres_fields is not None and isinstance(field, postgres_fields.ArrayField):
            return [cls.convert(field.base_field)]

        for model_type, primitive_type in cls.TYPES.items():
            if isinstance(field, model_type):
                return primitive_type

        raise ConverterException(f'Type "{field.__class__}" is unknown')


class FilterOperator(BaseFilterOperator):
    OPERATORS = {
        # operator:  (lookup_expr, negate needed)
        Operator.EQUAL: ("__exact", False),
        Operator.NOT_EQUAL: ("__exact", True),
        Operator.CONTAINS: ("__icontains", False),
        Operator.NOT_CONTAINS: ("__icontains", True),
        Operator.STARTS_WITH: ("__istartswith", False),
        Operator.ENDS_WITH: ("__iendswith", False),
        Operator.GREATER_THAN: ("__gt", False),
        Operator.AFTER: ("__gt", False),
        Operator.LESS_THAN: ("__lt", False),
        Operator.BEFORE: ("__lt", False),
        Operator.MISSING: ("__isnull", False),
        Operator.PRESENT: ("__isnull", True),
        Operator.IN: ("__in", False),
        Operator.NOT_IN: ("__in", True),
        Operator.INCLUDES_ALL: ("__contains", False),
        Operator.MATCH: ("__regex", False),
    }

    @classmethod
    def get_operator(cls, operator: Operator) -> Tuple[str, bool]:
        """return (expression_lookup, negate)"""
        try:
            return cls.OPERATORS[operator]
        except KeyError:
            raise ConverterException(f"Unable to handle the operator {operator}")

    @classmethod
    def get_operators_for_field(cls, field: models.Field) -> Set[Operator]:
        _type = TypeConverter.convert(field)
        wanted_operators = cls.get_for_type(_type)
        lookup_names = field.get_lookups().keys()
        unavailable_operators = []
        for wanted_operator in wanted_operators:
            if cls.OPERATORS.get(wanted_operator, ("",))[0].replace("__", "") not in lookup_names:
                unavailable_operators.append(wanted_operator)

        return set([operator for operator in wanted_operators if operator not in unavailable_operators])
